#include "stdafx.h"
#include "detectplate.h"
#include "dump.h"
#include "integralhistogram.h"
#include "profiling.h"
#include "prune_subwindows.h"

#include <sstream>

using namespace std;

pair<CvMat*, CvMat*> ClassifyImagePixels(CvMat * source_image,
                                         CvMat * cluster_means_norm,
                                         PixClass pixclass)
{
    CvMat * timg = cvCreateMat(source_image->rows, source_image->cols, CV_16SC1);
    CvMat * wimg = cvCreateMat(source_image->rows, source_image->cols, CV_32FC1);
    ilClassifyImagePixels(source_image, timg, wimg, cluster_means_norm, (int)pixclass);
    return make_pair(timg, wimg);
}


CvMat ** IntegralBlockHist(CvMat * timg, CvMat * wimg,
                           size_t cluster_number,
                           size_t dx, size_t dy)
{
    size_t nby = (div(timg->rows, dy)).quot;
    size_t nbx = (div(timg->cols, dx)).quot;
    CvMat ** iblockhist = new CvMat*[cluster_number];
    for (size_t i = 0; i < cluster_number; i++)
        iblockhist[i] = cvCreateMat(nby, nbx, CV_32FC1);

//    ilIntegralBlockHist2( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("ilIntegralBlockHist2");

    ilIntegralBlockHist( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("IntegralBlockHist");
    return iblockhist;
}

IMPEXP
SubWindows DetectPlate( CvMat * source_image,
                        Model const & model, 
                        CvMat * cluster_means,
                        CvMat * cov_matrix,
                        CvMat * cluster_means_norm )
{
    using namespace std;

    //*** label image pixels
    PixClass pixclass = TEXTONS_SC12_GDORDER2;
    // PixClass pixclass = ORIENTGRAD_DNUM4;

    /**
     * array subwindows passing the classification tests.
     */
    SubWindows good_windows;
    uint best_stage = 0;

    // we're going to scale down image and pass the detection test on it
    float scale_total = 1.;
    float const scale_step = .9;

    for (; scale_total > 0.4; scale_total *= scale_step)
	//scale_total=1.;
    {
        float const inverse_scale_total = 1 / scale_total;
        CvMat * scaled_image = NULL;
        if (scale_total < 1.)
        {
            uint scaled_x = (uint)(source_image->width * scale_total);
            uint scaled_y = (uint)(source_image->height * scale_total);
            scaled_image = cvCreateMat(scaled_y, scaled_x, CV_32F);
            cvResize( source_image, scaled_image, CV_INTER_CUBIC );
        }
        else
            scaled_image = source_image;

        pair<CvMat*, CvMat*> result =
                ClassifyImagePixels(scaled_image, cluster_means_norm, pixclass);

        if (scaled_image != source_image)
            cvReleaseMat( &scaled_image );

        CvMat * timg = result.first;
        CvMat * weighted_img = result.second;
//        DumpMatrix("../../dump/timg2.xml", timg);
//        DumpMatrix("../../dump/wimg2.xml", weighted_img);
        Tick("ClassifyImagePixels");
    
        size_t cluster_number = (pixclass == ORIENTGRAD_DNUM4) ? 4 : 12;
    
        // compute integral weighted block histogram
        uint block_size_x = 5;
        uint block_size_y = 5;
    
        CvMat ** integral_block_histogram =
                    IntegralBlockHist(timg,
                                      (pixclass == ORIENTGRAD_DNUM4) ? weighted_img
                                                                     : NULL,
                                      cluster_number, block_size_x, block_size_y);
    
        cvReleaseMat( &timg );
        cvReleaseMat( &weighted_img );
    
        // cache histogram size
        uint hist_size_x = integral_block_histogram[0]->cols;
        uint hist_size_y = integral_block_histogram[0]->rows;
    
        // the size of the subwindows
        uint subwin_size_x = 18;
        uint subwin_size_y = 5;
    
    
        // move wubwindow inside the hist-image with step (subwin_step_x, subwin_step_y).
        uint subwin_step_x = 1;
        uint subwin_step_y = 1;
    
        Tick("before AdaCascadeClass");
        /**
         * Note: we're not going to substruct the subwindow from hist-image and pass it to
         * the function-classifier, instead we're gonna pass the rectangle coordinates and
         * a reference to the image data.
         *
         * Original algorithm manipulated with large matrices representing all possible
         * subwindows. We're gonna call classifier on every subwindow in a cycle instead of
         * calling classifier only once and passing it a huge amount of data.
         */
        uint top, left;
        uint left_max = hist_size_x - subwin_size_x;     assert( left_max > 0 );
        uint top_max  = hist_size_y - subwin_size_y;     assert( top_max > 0 );
        ClearModelCache(model);
        for(left = 0; left < left_max; left += subwin_step_x)
        {
            for(top = 0; top < top_max; top += subwin_step_y)
            {
                // try to figure out if that region of the hist-image represents the Goal
                SubWindowInfo result =
                    adacascadeclass(model, integral_block_histogram,
                                    cluster_number, left, top, subwin_size_x, subwin_size_y);
                result.x *= (uint)(block_size_x * inverse_scale_total);
                result.y *= (uint)(block_size_y * inverse_scale_total);
                result.w *= (uint)(block_size_x * inverse_scale_total);
                result.h *= (uint)(block_size_y * inverse_scale_total);
                uint const stage = result.passed ? model.size() : result.stage;
                if (stage >= best_stage)
                {
                    if (stage > best_stage)
                    {
                        best_stage = stage;
                        good_windows.clear();
                    }
                    good_windows.push_back( result );
                }
            }
        }
        Tick("AdaCascadeClass");
    
        for (size_t i = 0; i < cluster_number; i++)
            cvReleaseMat( &integral_block_histogram[i] );
        delete[] integral_block_histogram;
    }
//    DumpResultingWindows( basepath + "dump/resulting_windows.xml", good_windows );

	return good_windows;

#if 0 //*** pruning
    SubWindows subwindow_clusters;
    PruneSubWindows( good_windows, 0.3, 1, subwindow_clusters );
    if (subwindow_clusters.empty())
        subwindow_clusters = good_windows;
    Tick("PruneSubWindows");

    for (SubWindows::const_iterator iit = subwindow_clusters.begin();
                                    iit != subwindow_clusters.end(); ++iit)
    {
        std::cerr << "--> Window(" << iit->x << "," << iit->y << "," << iit->w << "," << iit->h << ";" << iit->cluster_size << ")\n" << std::flush;
    }

//    DumpResultingWindows( basepath + "dump/resulting_pruned_windows.xml", subwindow_clusters );
    Tick("SubWindow clusters dumped");

    return subwindow_clusters;
#endif
}

void ClearModelCache(Model const & model)
{
    for (Model::const_iterator it = model.begin(),
                              end = model.end();
                              it != end; ++it)
    {
        for (StageRules::const_iterator rit = it->rules.begin(),
                                       rend = it->rules.end();
                                       rit != rend; ++rit)
        {
            rit->clear_cache();
        }
    }
}

